source("dataS/scripts/vehicles/specializations/events/ReceivingHopperSetCreateBoxesEvent.lua")

BaggingMachine = {
	prerequisitesPresent = function (specializations)
		return SpecializationUtil.hasSpecialization(FillUnit, specializations) and SpecializationUtil.hasSpecialization(Dischargeable, specializations)
	end,
	initSpecialization = function ()
		local schema = Vehicle.xmlSchema

		schema:setXMLSpecializationType("BaggingMachine")
		schema:register(XMLValueType.INT, "vehicle.baggingMachine#fillUnitIndex", "Fill unit index", 1)
		schema:register(XMLValueType.STRING, "vehicle.baggingMachine#dependsOnAnimation", "Needs Animation xy for enabling spawner", "false")
		schema:register(XMLValueType.NODE_INDEX, "vehicle.baggingMachine.boxes#spawnPlaceNode", "Spawn place node")
		schema:register(XMLValueType.STRING, "vehicle.baggingMachine.boxes.box(?)#fillType", "Fill type name")
		schema:register(XMLValueType.STRING, "vehicle.baggingMachine.boxes.box(?)#filename", "Box filename")
		schema:setXMLSpecializationType()

		local schemaSavegame = Vehicle.xmlSchemaSavegame

		schemaSavegame:register(XMLValueType.BOOL, "vehicles.vehicle(?).FS22_lsfmBaggingPack.baggingMachine#createBoxes", "Create boxes")
	end
}

BaggingMachine.MOD_NAME = g_currentModName or ""
BaggingMachine.SPEC_NAME = string.format("spec_%s.baggingMachine", g_currentModName)

function BaggingMachine.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "setCreateBoxes", BaggingMachine.setCreateBoxes)
	SpecializationUtil.registerFunction(vehicleType, "getCanSpawnNextBox", BaggingMachine.getCanSpawnNextBox)
	SpecializationUtil.registerFunction(vehicleType, "collisionTestCallback", BaggingMachine.collisionTestCallback)
	SpecializationUtil.registerFunction(vehicleType, "createBox", BaggingMachine.createBox)
	SpecializationUtil.registerFunction(vehicleType, "onCreateBoxFinished", BaggingMachine.onCreateBoxFinished)
end

function BaggingMachine.registerOverwrittenFunctions(vehicleType)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "handleDischargeRaycast", BaggingMachine.handleDischargeRaycast)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanBeSelected", BaggingMachine.getCanBeSelected)
end

function BaggingMachine.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onLoad", BaggingMachine)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", BaggingMachine)
	SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", BaggingMachine)
	SpecializationUtil.registerEventListener(vehicleType, "onFillUnitFillLevelChanged", BaggingMachine)
end

function BaggingMachine:onLoad(savegame)
	self.spec_baggingMachine = self[BaggingMachine.SPEC_NAME]
	local spec = self.spec_baggingMachine

	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine#unloadingDelay", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine#unloadInfoIndex", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine#dischargeInfoIndex", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.tipTrigger#index", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.boxTrigger#index", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.fillScrollerNodes.fillScrollerNode", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.fillEffect", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.fillEffect", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.boxTrigger#litersPerMinute", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.raycastNode#index", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.raycastNode#raycastLength", "Dischargeable functionalities")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.boxTrigger#boxSpawnPlaceIndex", "vehicle.baggingMachine.boxes#spawnPlaceNode")
	XMLUtil.checkDeprecatedXMLElements(self.xmlFile, "vehicle.baggingMachine.boxTrigger.box(0)", "vehicle.baggingMachine.boxes.box(0)")

	spec.fillUnitIndex = self.xmlFile:getValue("vehicle.baggingMachine#fillUnitIndex", 1)
	spec.dependsOnAnimation = self.xmlFile:getValue("vehicle.baggingMachine#dependsOnAnimation", "false")
	spec.spawnPlace = self.xmlFile:getValue("vehicle.baggingMachine.boxes#spawnPlaceNode", nil, self.components, self.i3dMappings)
	spec.boxes = {}
	local i = 0

	while true do
		local baseName = string.format("vehicle.baggingMachine.boxes.box(%d)", i)

		if not self.xmlFile:hasProperty(baseName) then
			break
		end

		local fillTypeStr = self.xmlFile:getValue(baseName .. "#fillType")
		local filename = self.xmlFile:getValue(baseName .. "#filename")
		local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(fillTypeStr)

		if fillTypeIndex ~= nil then
			spec.boxes[fillTypeIndex] = filename
		else
			Logging.xmlWarning(self.xmlFile, "Invalid fillType '%s'", fillTypeStr)
		end

		i = i + 1
	end

	spec.createBoxes = false
	spec.lastBox = nil
	spec.creatingBox = false

	if savegame ~= nil then
		spec.createBoxes = savegame.xmlFile:getValue(savegame.key .. ".FS22_lsfmBaggingPack.baggingMachine#createBoxes", spec.createBoxes)
	end

	if not self.isServer then
		SpecializationUtil.removeEventListener(self, "onUpdateTick", BaggingMachine)
	end
end

function BaggingMachine:saveToXMLFile(xmlFile, key, usedModNames)
	local spec = self.spec_baggingMachine

	xmlFile:setValue(key .. "#createBoxes", spec.createBoxes)
end

function BaggingMachine:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local spec = self.spec_baggingMachine

	if spec.createBoxes and self:getDischargeState() == Dischargeable.DISCHARGE_STATE_OFF and self:getCanSpawnNextBox() then
		self:createBox()
	end
end

function BaggingMachine:setCreateBoxes(state, noEventSend)
	local spec = self.spec_baggingMachine

	if state ~= spec.createBoxes then
		ReceivingHopperSetCreateBoxesEvent.sendEvent(self, state, noEventSend)

		spec.createBoxes = state

		BaggingMachine.updateActionEvents(self)

		spec.lastBox = nil
	end
end

function BaggingMachine:getCanSpawnNextBox()
	local spec = self.spec_baggingMachine

	if spec.dependsOnAnimation ~= nil and spec.dependsOnAnimation ~= "false" then
		local animation = self:getAnimationByName(spec.dependsOnAnimation)

		if animation ~= nil then
			local isAnimationPlaying = self:getIsAnimationPlaying(spec.dependsOnAnimation)
			local animationTime = self:getAnimationTime(spec.dependsOnAnimation)
			local stopTime = animation.stopTime

			--print("Animation: " .. tostring(spec.dependsOnAnimation))
			--print("Is playing: " .. tostring(isAnimationPlaying))
			--print("Animation time: " .. tostring(animationTime))

			if animationTime ~= 0 then
				if spec.lastBox ~= nil and spec.lastBox:getFillUnitFreeCapacity(1) > 0 then
					spec.lastBox = nil
				end

				--print("Returning false because animation time ~= 0")
				return false
			end
		end
	end

	if spec.creatingBox then
		--print("Returning false because creating box")
		return false
	end

	local fillType = self:getFillUnitFillType(spec.fillUnitIndex)

	if spec.boxes[fillType] ~= nil then
		if spec.lastBox ~= nil and spec.lastBox:getFillUnitFreeCapacity(1) > 0 then
			--print("Returning false because last box has free capacity")
			return false
		end

		local xmlFilename = Utils.getFilename(spec.boxes[fillType], self.baseDirectory)
		local size = StoreItemUtil.getSizeValues(xmlFilename, "vehicle", 0)
		local x, y, z = getWorldTranslation(spec.spawnPlace)
		local rx, ry, rz = getWorldRotation(spec.spawnPlace)
		spec.foundObjectAtSpawnPlace = false

		overlapBox(x, y, z, rx, ry, rz, size.width * 0.5, 2, size.length * 0.5, "collisionTestCallback", self, 5468288)

		--print("Returning" .. tostring(not spec.foundObjectAtSpawnPlace) .. " because found object at spawn place")
		return not spec.foundObjectAtSpawnPlace
	end

	--print("Returning false because no box for fill type")
	return false
end

function BaggingMachine:collisionTestCallback(transformId)
	if (g_currentMission.nodeToObject[transformId] ~= nil or g_currentMission.players[transformId] ~= nil) and g_currentMission.nodeToObject[transformId] ~= self then
		local spec = self.spec_baggingMachine
		spec.foundObjectAtSpawnPlace = true
	end
end

function BaggingMachine:createBox()
	local spec = self.spec_baggingMachine

	if self.isServer and spec.createBoxes then
		local fillType = self:getFillUnitFillType(spec.fillUnitIndex)

		if spec.boxes[fillType] ~= nil then
			local x, _, z = getWorldTranslation(spec.spawnPlace)
			local dirX, _, dirZ = localDirectionToWorld(spec.spawnPlace, 0, 0, 1)
			local yRot = MathUtil.getYRotationFromDirection(dirX, dirZ)
			local xmlFilename = Utils.getFilename(spec.boxes[fillType], self.baseDirectory)
			local location = {
				x = x,
				z = z,
			    yRot = yRot
			}

			spec.creatingBox = true
			
			location.yOffset = 0.53

			VehicleLoadingUtil.loadVehicle(xmlFilename, location, true, 0, Vehicle.PROPERTY_STATE_OWNED, self:getOwnerFarmId(), nil, nil, self.onCreateBoxFinished, self)
		end
	end
end

function BaggingMachine:onCreateBoxFinished(vehicle, vehicleLoadState, arguments)
	local spec = self.spec_baggingMachine
	spec.creatingBox = false

	if vehicleLoadState == VehicleLoadingUtil.VEHICLE_LOAD_OK then
		spec.lastBox = vehicle
	end
end

function BaggingMachine:handleDischargeRaycast(superFunc, dischargeNode, hitObject, hitShape, hitDistance, hitFillUnitIndex, hitTerrain)
	local stopDischarge = false

	if hitObject ~= nil then
		local fillType = self:getDischargeFillType(dischargeNode)
		local allowFillType = hitObject:getFillUnitAllowsFillType(hitFillUnitIndex, fillType)

		if allowFillType and hitObject:getFillUnitFreeCapacity(hitFillUnitIndex) > 0 then
			self:setDischargeState(Dischargeable.DISCHARGE_STATE_OBJECT, true)
		else
			stopDischarge = true
		end
	else
		stopDischarge = true
	end

	if stopDischarge and self:getDischargeState() == Dischargeable.DISCHARGE_STATE_OBJECT then
		self:setDischargeState(Dischargeable.DISCHARGE_STATE_OFF, true)
	end
end

function BaggingMachine:getCanBeSelected(superFunc)
	return true
end

function BaggingMachine:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	if self.isClient then
		local spec = self.spec_baggingMachine

		self:clearActionEventsTable(spec.actionEvents)

		if isActiveForInputIgnoreSelection then
			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.IMPLEMENT_EXTRA, self, BaggingMachine.actionEventToggleBoxCreation, false, true, false, true, nil)

			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL)
			BaggingMachine.updateActionEvents(self)
		end
	end
end

function BaggingMachine:onFillUnitFillLevelChanged(fillUnitIndex, fillLevelDelta, fillType, toolType, fillPositionData, appliedDelta)
	if self:getFillUnitFillLevel(fillUnitIndex) > 0 then
		self:raiseActive()
	end
end

function BaggingMachine:actionEventToggleBoxCreation(actionName, inputValue, callbackState, isAnalog)
	local spec = self.spec_baggingMachine

	self:setCreateBoxes(not spec.createBoxes)
end

function BaggingMachine:updateActionEvents()
	local spec = self.spec_baggingMachine
	local actionEvent = spec.actionEvents[InputAction.IMPLEMENT_EXTRA]

	if actionEvent ~= nil then
		if spec.createBoxes then
			g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("action_disableBagSpawning"))
		else
			g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("action_enableBagSpawning"))
		end
	end
end
